#include "cylinder.h"
#include "misc.h"
#include <cmath>

Cylinder::Cylinder(float radius, float length, Vector3D bottom, Vector3D xAxis, Vector3D yAxis, Vector3D zAxis, RGBColour *texture, int textureWidth, int textureHeight, int specularExponent, float specularReflectivity)
:Primitive(texture, textureWidth, textureHeight, specularExponent, specularReflectivity)
{
   this->radius = radius;
   this->length = length;
   this->xAxis = xAxis;
   this->yAxis = yAxis;
   this->zAxis = zAxis;
   this->bottom = bottom;
}

float Cylinder::get_t(Ray ray)
{
   float xComponent = ray.direction.dot(xAxis);
   float yComponent = ray.direction.dot(yAxis);
   // Below the variables a, b, c are the coefficients of a quadratic in t with solutions t1,2.
   float a = xComponent * xComponent + yComponent * yComponent;
   if(!a) return nointersection_t_Value; // No intersection because the ray is parallel to the principle axis of the cylinder.
   float tempX = (ray.source - bottom).dot(xAxis), tempY = (ray.source - bottom).dot(yAxis);
   float b = 2 * (xComponent * tempX + yComponent * tempY);
   float c = tempX * tempX + tempY * tempY - radius * radius;
   float discriminant = b * b - 4 * a * c;
   if(discriminant < 0) return nointersection_t_Value; // No intersection because the ray just passes by the cylinder.
   float t, t1 = (- b + sqrt(discriminant)) / (2 * a), t2 = - b / a - t1; // NB: The sum of the roots of a quadratic is - b / a etc..
   float zComponent1 = (ray.source + ray.direction * t1 - bottom).dot(zAxis);
   float zComponent2 = (ray.source + ray.direction * t2 - bottom).dot(zAxis);
   if(t1 < 0 || zComponent1 < 0 || zComponent1 > length) t = t2;
   else if(t2 < 0 || zComponent2 < 0 || zComponent2 > length) t = t1;
   else t = min(t1, t2);
   Vector3D intersection = ray.source + ray.direction * t;
   float zComponent = (intersection - bottom).dot(zAxis);
   if(t < 0 || zComponent < 0 || zComponent > length) return nointersection_t_Value; // No intersection in front of the ray source and inside the cylinder boundary.
   return t;
}

RGBColour Cylinder::getColour(Vector3D intersection)
{
   Vector3D tempVect = intersection - bottom;
   int u = round(iTan(tempVect.dot(xAxis), tempVect.dot(yAxis)) * (textureWidth - 1) / (2 * PI));
   int v = round(tempVect.dot(zAxis) * (textureHeight - 1) / length);
   return texture[u + v * textureWidth];
}

Vector3D Cylinder::getNormal(Vector3D intersection)
{
   Vector3D tempVect = intersection - bottom;
   Vector3D normal = tempVect - zAxis * tempVect.dot(zAxis);
   normal.normalise();
   return normal;
}
