#include "ellipsoid.h"
#include "misc.h"
#include <cmath>

Ellipsoid::Ellipsoid(float xRadius, float yRadius, float zRadius, Vector3D centre, Matrix3x3 transitionMatrix, RGBColour *texture, int textureWidth, int textureHeight, int specularExponent, float specularReflectivity)
:Primitive(texture, textureWidth, textureHeight, specularExponent, specularReflectivity)
{
   this->xRadius = xRadius;
   this->yRadius = yRadius;
   this->zRadius = zRadius;
   this->centre = centre;
   this->transitionMatrix = transitionMatrix;
   transitionMatrixTranspose = transitionMatrix.transpose();
}

float Ellipsoid::get_t(Ray ray)
{
   Vector3D tempVect = transitionMatrix * (ray.source - centre);
   Vector3D newDirection = transitionMatrix * ray.direction;
   float a = newDirection.x * newDirection.x / (xRadius * xRadius) +
             newDirection.y * newDirection.y / (yRadius * yRadius) +
             newDirection.z * newDirection.z / (zRadius * zRadius);
   float b = 2 * (newDirection.x * tempVect.x / (xRadius * xRadius) +
                  newDirection.y * tempVect.y / (yRadius * yRadius) +
                  newDirection.z * tempVect.z / (zRadius * zRadius));
   float c = tempVect.x * tempVect.x / (xRadius * xRadius) +
             tempVect.y * tempVect.y / (yRadius * yRadius) +
             tempVect.z * tempVect.z / (zRadius * zRadius) - 1;
   float discriminant = b * b - 4 * a * c;
   if(discriminant < 0) return nointersection_t_Value; // No intersection because the ray just passes by the ellipsoid.
   // I haven't coded the below bit, which decides which point of intersection to use, the most efficient way at all, but the below method is easy to follow and is only slightly slower than necessary.
   float t, t1 = (- b + sqrt(discriminant)) / (2 * a), t2 = - b / a - t1; // NB: The sum of the roots of a quadratic is - b / a etc..
   if(t1 > 0)
   {
      if(t2 > 0) t = min(t1, t2);
      else t = t1;
   }
   else t = t2;
   if(t < 0) return nointersection_t_Value;
   return t;
}

RGBColour Ellipsoid::getColour(Vector3D intersection)
{
   Vector3D tempVect = transitionMatrix * (intersection - centre);
   float theta, phi;
   theta = iTan(tempVect.z / zRadius, sqrt(tempVect.x * tempVect.x / (xRadius * xRadius) + tempVect.y * tempVect.y / (yRadius * yRadius)));
   phi = iTan(tempVect.x / xRadius, tempVect.y / yRadius);
   int u = round((textureWidth - 1) * theta / PI);
   int v = round((textureHeight - 1) * phi / (2 * PI));
   return texture[u + v * textureWidth];
}

Vector3D Ellipsoid::getNormal(Vector3D intersection)
{
   Vector3D normal = transitionMatrix * (intersection - centre);
   normal.x /= (xRadius * xRadius);
   normal.y /= (yRadius * yRadius);
   normal.z /= (zRadius * zRadius);
   normal = transitionMatrixTranspose * normal;
   normal.normalise();
   return normal;
}
