#include "scene.h"
#include "misc.h"
#include <cmath>

Scene::Scene(RGBColour ambientLight)
{
   this->ambientLight = ambientLight;
   nLightSources = 0;
   nPrimitives = 0;
   recursionLevel = 0;
}

int Scene::addPrimitive(Primitive *p)
{
   if(nPrimitives < maxPrimitives)
   {
      primitives[nPrimitives++] = p;
      return 1;
   }
   return 0;
}

int Scene::addLightSource(LightSource *l)
{
   if(nLightSources < maxLightSources)
   {
      lightSources[nLightSources++] = l;
      return 1;
   }
   return 0;
}

float Scene::get_t(Ray ray, Primitive*& Obj)
{
   float current_t, new_t;
   int i, j;

   current_t = primitives[0]->get_t(ray);
   j = 0;
   for(i = 0; i < nPrimitives; i++)
   {
      new_t = primitives[i]->get_t(ray);
      if(new_t > 0 && (new_t < current_t || current_t < 0))
      {
         current_t = new_t;
         j = i;
      }
   }
   Obj = primitives[j];
   return current_t;
}

RGBColour Scene::doLighting(Vector3D intersection, Vector3D viewDirection, Primitive &Obj)
{
   // Calculate the ambient light.
   RGBColour ambient, directLocalLight;

   ambient = Obj.getColour(intersection) * ambientLight * Obj.diffuseReflectivity;

   // Get the direct local light for this object.
   int i;
   for(i = 0; i < nLightSources; i++)
   {
      if(isOccluded(intersection, *lightSources[i])) continue; // Creates shadows.
      directLocalLight += Obj.getLocalLight(intersection, viewDirection, *(lightSources[i]));
   }

   // Calculate the specular light reflected from the other objects.
   RGBColour reflectedSpecularLight;

   if(recursionLevel < maxRecursionLevel)
   {
      Primitive *Obj2;
      Vector3D normal = Obj.getNormal(intersection);
      Vector3D reflectedDirection = viewDirection - normal * 2 * normal.dot(viewDirection); // This vector is actually also calculated in the getLocalLight method of the Primitive class.
      reflectedDirection.normalise();
      Ray reflectedRay(intersection + reflectedDirection * 0.05f, reflectedDirection);
      float t;

      if((t = get_t(reflectedRay, Obj2)) > 0)
      {
         recursionLevel++;
         reflectedSpecularLight = doLighting(reflectedRay.source + reflectedRay.direction * t, reflectedRay.direction, *Obj2);
         recursionLevel--;
      }
   }
   return ambient + directLocalLight + reflectedSpecularLight * Obj.specularReflectivity;
}

int Scene::isOccluded(Vector3D p, LightSource l)
{
   int i;
   Ray r(p + l.direction * -0.05f, l.direction * -1);
   for(i = 0; i < nPrimitives; i++)
      if(primitives[i]->get_t(r) > 0.0f) break;
   if(i == nPrimitives) return 0;
   return 1;
}
